require "defines"
require "util"

-- EARTHQUAKE SCRIPTS
local quake_start = {}
local quake_end = {}
local quake_power = {}

local quake_power_to_distance = 160
local quake_power_to_duration = 120

-- do not set power above 3...
function Quake(player, power)
	local s = game.surfaces["nauvis"]
	local move = {x = math.random() * math.random() * power, y = math.random() * math.random() * power}
	local result = player.position
	if math.random() < 0.5 then
		move.x = -move.x
		end
	if math.random() < 0.5 then
		move.y = -move.y
		end
	if player.vehicle ~= nil then
--[[		local vehicle = player.vehicle
		result.x = result.x + (move.x/2)
		result.y = result.y + (move.y/2)
		if s.can_place_entity{name = vehicle.name, position = result} then
			vehicle.teleport(result)
			end]]--
		-- move the vehicle? haven't got this working in a way that isn't game-breaking yet
	else
		result.x = result.x + move.x
		result.y = result.y + move.y
		if s.can_place_entity{name = "player", position = result} then
			player.teleport(result)
			end
		end
	end

function Quake_Update(player_index)
	local intensity = quake_power[player_index]
	local duration = quake_end[player_index] - quake_start[player_index]
	local runoff = (game.tick - quake_start[player_index]) / 2
	local power = intensity * (duration - runoff) / duration
	Quake(game.get_player(player_index),power)
	end

function Quake_Begin(player_index, power, duration)
	local current_quake
	if quake_power[player_index] ~= nil then current_quake = quake_power[player_index] else current_quake = 0 end
	if (power > current_quake) and (duration > 5) then
		quake_start[player_index] = game.tick
		quake_end[player_index] = game.tick + duration
		quake_power[player_index] = power
	else
		Quake(game.get_player(player_index),power)
		end
	end

function Quake_Area(position, power)
	for i,player in pairs(game.players) do
		local p = player.index
		local XY = player.position
		local distance = 1000
		local scaled = power
		if position ~= nil and XY ~= nil then
			distance = util.distance(position, XY)
			end
		power = power - (distance/quake_power_to_distance)
		if scaled > 0 then
			local s = game.surfaces["nauvis"]
			if scaled > 2 then
				s.create_entity{name = "worm-rumble-large", position = XY}
			elseif scaled > 1 then
				s.create_entity{name = "worm-rumble-medium", position = XY}
			else
				s.create_entity{name = "worm-rumble-small", position = XY}
				end
			Quake_Begin (p, scaled, (scaled+power) * quake_power_to_duration / 2)
			end
		end
	end

function check_quakes()
	if game.tick % 5 == 0 then
		for i,player in pairs(game.players) do
			local p = player.index
			if quake_end[p] ~= nil then
				if quake_end[p] > game.tick then
					Quake_Update(p)
					end
				end
			end
		end
	end
	
-- WORM ATTACK SCRIPTS
--[[
Worms evolve like spawners, going from small to medium to big to behemoth... to colossal
Once they have reached behemoth size there is a chance that they will burrow and go in search of the player
Once they have reached colossal size they will automatically burrow on the next update tick
When a worm burrows it leaves behind small worms
Worms will not burrow while close to the player or while within range of any player units
Evolving and choosing to burrow get updated once every 5 minutes
Worms only evolve or burrow if in contact with any pollution, and if between a distance of 150 and 1200 of the player
(the bottom end to avoid changing it while on screen, the pollution and the top end to avoid needing too much processing)

Once burrowed, each update tick (30 seconds) the worms that are burrowed will either surface to attack or move to the next chunk
They will surface and attack if they are in the most polluted nearby chunk and there are any player units in the chunk
They will move to the most polluted adjacent chunk if there are no player units in their chunk
If there are player units in their chunk, but an adjacent chunk is more polluted, it will choose based on the health of player units and the difference in pollution levels
(it will ignore small numbers of player units if the pollution difference is big enough)

Once in attack mode they will launch an area attack every second to make space until they have room to surface, then they will surface and resume normal attacks
]]--
local worm_evo_medium = 0.5
local worm_evo_big = 0.8
local worm_evo_behemoth = 0.95
local worm_evo_colossal = 0.99
local worm_evo_min_distance = 150
local worm_evo_max_distance = 1200
local worm_attack_behemoth_chance = 0.25
local worm_attack_colossal_chance = 1.0
local worm_evolve_update_ticks = 5*60*60
local worm_burrow_update_ticks = 30*60
local worm_attack_update_ticks = 1*60
local worm_pollution_antagonise = 0
local worm_pollution_weight = 10
local worm_attacking_xoffset = -10
local worm_attacking_yoffset = 0

-- find all the worms near the player and in non-zero pollution
function find_worms ()
	local surface = game.surfaces["nauvis"]
	global.surface_worms = {}
	for c in surface.get_chunks() do
		local longest = worm_evo_max_distance
		local XY = {x=c.x * 32 + 16, y=c.y * 32 + 16}
		for i,player in pairs(game.players) do
			local distance = worm_evo_max_distance
			if player.position ~= nil then distance = util.distance(XY,player.position) end
			if distance < longest then longest = distance end
			end
		if surface.get_pollution(XY) > worm_pollution_antagonise 
			and longest < worm_evo_max_distance
			and longest > worm_evo_min_distance then
			for key, surface_worms in pairs(surface.find_entities_filtered({area={{c.x * 32, c.y * 32}, {c.x * 32 + 32, c.y * 32 + 32}}, name="small-worm-turret"})) do
				table.insert(global.surface_worms, surface_worms)
				end
			for key, surface_worms in pairs(surface.find_entities_filtered({area={{c.x * 32, c.y * 32}, {c.x * 32 + 32, c.y * 32 + 32}}, name="medium-worm-turret"})) do
				table.insert(global.surface_worms, surface_worms)
				end
			for key, surface_worms in pairs(surface.find_entities_filtered({area={{c.x * 32, c.y * 32}, {c.x * 32 + 32, c.y * 32 + 32}}, name="big-worm-turret"})) do
				table.insert(global.surface_worms, surface_worms)
				end
			for key, surface_worms in pairs(surface.find_entities_filtered({area={{c.x * 32, c.y * 32}, {c.x * 32 + 32, c.y * 32 + 32}}, name="behemoth-worm-turret"})) do
				table.insert(global.surface_worms, surface_worms)
				end
			for key, surface_worms in pairs(surface.find_entities_filtered({area={{c.x * 32, c.y * 32}, {c.x * 32 + 32, c.y * 32 + 32}}, name="colossal-worm-turret"})) do
				table.insert(global.surface_worms, surface_worms)
				end
			end
		end
	end

-- evolve if appropriate
function check_worm_evolution ()
	local surface = game.surfaces["nauvis"]
    if global.surface_worms ~= nil then
		for k,worm in pairs(global.surface_worms) do
			if worm.valid then
				local X = worm.position.x
				local Y = worm.position.y
				if worm.name == "small-worm-turret" then
					local excess = game.evolution_factor - worm_evo_medium
					local upgrade = excess / (worm_evo_big - worm_evo_medium) / 2
					-- so if the evolution has reached the next tier, there is a 50% chance of upgrade each upgrade tick
					if math.random() < upgrade then
						local X2 = X + math.random(-5,5)
						local Y2 = Y + math.random(-5,5)
						local evotype = "medium-worm-turret"

						worm.destroy()

--						game.player.print("Grow!")
						if surface.can_place_entity{name = evotype, position = {x=X2,y=Y2}} then
							surface.create_entity{name = evotype, position = {x=X2,y=Y2}, force= game.forces.enemy}
						elseif surface.can_place_entity{name = evotype, position = {x=X,y=Y}} then
							surface.create_entity{name = evotype, position = {x=X,y=Y}, force= game.forces.enemy}
						else
							surface.create_entity{name = "small-worm-turret", position = {x=X,y=Y}, force= game.forces.enemy}
							end
						end
				elseif worm.name == "medium-worm-turret" then
					local excess = game.evolution_factor - worm_evo_big
					local upgrade = excess / (worm_evo_behemoth - worm_evo_big) / 2
					-- so if the evolution has reached the next tier, there is a 50% chance of upgrade each upgrade tick
					if math.random() < upgrade then
						local X2 = X + math.random(-5,5)
						local Y2 = Y + math.random(-5,5)
						local evotype = "big-worm-turret"

						worm.destroy()

--						game.player.print("Grow!")
						if surface.can_place_entity{name = evotype, position = {x=X2,y=Y2}} then
							surface.create_entity{name = evotype, position = {x=X2,y=Y2}, force= game.forces.enemy}
						elseif surface.can_place_entity{name = evotype, position = {x=X,y=Y}} then
							surface.create_entity{name = evotype, position = {x=X,y=Y}, force= game.forces.enemy}
						else
							surface.create_entity{name = "medium-worm-turret", position = {x=X,y=Y}, force= game.forces.enemy}
							end
						end
				elseif worm.name == "big-worm-turret" then
					local excess = game.evolution_factor - worm_evo_behemoth
					local upgrade = excess / (worm_evo_colossal - worm_evo_behemoth) / 2
					-- so if the evolution has reached the next tier, there is a 50% chance of upgrade each upgrade tick
					if math.random() < upgrade then
						local X2 = X + math.random(-5,5)
						local Y2 = Y + math.random(-5,5)
						local evotype = "behemoth-worm-turret"

						worm.destroy()

--						game.player.print("Grow!")
						if surface.can_place_entity{name = evotype, position = {x=X2,y=Y2}} then
							surface.create_entity{name = evotype, position = {x=X2,y=Y2}, force= game.forces.enemy}
						elseif surface.can_place_entity{name = evotype, position = {x=X,y=Y}} then
							surface.create_entity{name = evotype, position = {x=X,y=Y}, force= game.forces.enemy}
						else
							surface.create_entity{name = "big-worm-turret", position = {x=X,y=Y}, force= game.forces.enemy}
							end
						end
				elseif worm.name == "behemoth-worm-turret" then
					local excess = game.evolution_factor - worm_evo_colossal
					local upgrade = excess / (1 - worm_evo_colossal) / 2
					-- so if the evolution has reached the next tier, there is a 50% chance of upgrade each upgrade tick
					if math.random() < upgrade then
						local X2 = X + math.random(-5,5)
						local Y2 = Y + math.random(-5,5)
						local evotype = "colossal-worm-turret"

						worm.destroy()

--						game.player.print("Grow!")
						if surface.can_place_entity{name = evotype, position = {x=X2,y=Y2}} then
							surface.create_entity{name = evotype, position = {x=X2,y=Y2}, force= game.forces.enemy}
						elseif surface.can_place_entity{name = evotype, position = {x=X,y=Y}} then
							surface.create_entity{name = evotype, position = {x=X,y=Y}, force= game.forces.enemy}
						else
							surface.create_entity{name = "behemoth-worm-turret", position = {x=X,y=Y}, force= game.forces.enemy}
							end
						end
					end
				end
			end
		end
	end

-- check worms that are suitable to go hunting and possibly burrow them
function check_worm_burrow ()
	local surface = game.surfaces["nauvis"]
	if global.burrowing_worms == nil then
		global.burrowing_worms = {}
		end
    if global.surface_worms ~= nil then
		for k,worm in pairs(global.surface_worms) do
			if worm.valid then
				local X = worm.position.x
				local Y = worm.position.y
				if worm.name == "behemoth-worm-turret" then
					local excess = game.evolution_factor - worm_evo_behemoth
					local attack = worm_attack_behemoth_chance * excess / (worm_evo_colossal - worm_evo_behemoth)
					local enemies = false
					for i, targets in pairs(surface.find_entities_filtered({area={{X-30, Y-30}, {X+30, Y+30}}, force= game.forces.player})) do
						if util.distance(worm.position,targets.position) < 30 then
							if targets.health ~= nil then enemies = true end
							end
						end
					if enemies == true then attack = 0 end
					if math.random() < attack then
						--add behemoth worm to burrow table
						local X2 = X + math.random(-5,5)
						local Y2 = Y + math.random(-5,5)

						Quake_Area(worm.position, 2)
						worm.destroy()

						if surface.can_place_entity{name = "small-worm-turret", position = {x=X2,y=Y2}} then
							surface.create_entity{name = "small-worm-turret", position = {x=X2,y=Y2}, force= game.forces.enemy}
						else
							surface.create_entity{name = "small-worm-turret", position = {x=X,y=Y}, force= game.forces.enemy}
							end
						--table.insert(global.surface_worms, surface_worms)
						local burrowing_worm = surface.create_entity{name = "behemoth-worm-burrowing", position = {x=X,y=Y}, force= game.forces.neutral}
						table.insert(global.burrowing_worms,burrowing_worm)
--						game.player.print("Burrow")
						end
				elseif worm.name == "colossal-worm-turret" then
					-- always attack if not in combat
					--add colossal worm to burrow table
					local enemies = false
					for i, targets in pairs(surface.find_entities_filtered({area={{X-35, Y-35}, {X+35, Y+35}}, force= game.forces.player})) do
						if util.distance(worm.position,targets.position) < 35 then
							if targets.health ~= nil then enemies = true end
							end
						end
					if enemies == false then
						local X2 = X + math.random(-5,5)
						local Y2 = Y + math.random(-5,5)
						local X3 = X + math.random(-5,5)
						local Y3 = Y + math.random(-5,5)
						
						Quake_Area(worm.position, 3)
						worm.destroy()

						if surface.can_place_entity{name = "small-worm-turret", position = {x = X2,y = Y2}} then
							surface.create_entity{name = "small-worm-turret", position = {x = X2,y = Y2}, force= game.forces.enemy}
						elseif surface.can_place_entity{name = "small-worm-turret", position = {x = X3,y = Y3}} then
							surface.create_entity{name = "small-worm-turret", position = {x = X3,y = Y3}, force= game.forces.enemy}
							end
						if surface.can_place_entity{name = "small-worm-turret", position = {x = X3,y = Y3}} then
							surface.create_entity{name = "small-worm-turret", position = {x = X3,y = Y3}, force= game.forces.enemy}
						else
							surface.create_entity{name = "small-worm-turret", position = {x = X,y = Y}, force= game.forces.enemy}
							end
						local burrowing_worm = surface.create_entity{name = "colossal-worm-burrowing", position = {x = X,y = Y}, force= game.forces.neutral}
						table.insert(global.burrowing_worms,burrowing_worm)
--						game.player.print("Burrow")
						end
					end
				end
			end
		end
	end

-- check for burrowed worms, if in a suitable chunk they attack, otherwise they move
-- if they move to a cell which the player is in, cause a quake
function check_burrowing_worms ()
	local surface = game.surfaces["nauvis"]
	if global.burrowing_worms == nil then global.burrowing_worms = {} end
	if global.attacking_worms == nil then global.attacking_worms = {} end
	for i, burrowing_worms in pairs(global.burrowing_worms) do
		if burrowing_worms.valid then
			local position = burrowing_worms.position
			local chunk_mid = {x=position.x, y=position.y}
			local chunk_pol = 0
			local chunk_up_mid = {x=position.x, y=position.y - 32}
			local chunk_up_pol = 0
			local chunk_rt_mid = {x=position.x + 32, y=position.y}
			local chunk_rt_pol = 0
			local chunk_dn_mid = {x=position.x, y=position.y + 32}
			local chunk_dn_pol = 0
			local chunk_lt_mid = {x=position.x - 32, y=position.y}
			local chunk_lt_pol = 0
			local most_pol_loc = {x=0,y=0}
			local most_pol = 0
			local attack_type = ""
			local attack_endurance = 0
			if burrowing_worms.name == "behemoth-worm-burrowing" then
				attack_type = "behemoth-worm-attacking"
				attack_endurance = 10
			elseif burrowing_worms.name == "colossal-worm-burrowing" then
				attack_type = "colossal-worm-attacking"
				attack_endurance = 15
				end

			chunk_pol = surface.get_pollution(chunk_mid)
			most_pol_loc = chunk_mid
			most_pol = chunk_pol
			
			if surface.is_chunk_generated{math.floor(chunk_up_mid.x/32),math.floor(chunk_up_mid.y/32)} then
				chunk_up_pol = surface.get_pollution(chunk_up_mid)
				if chunk_up_pol > most_pol then
					most_pol = chunk_up_pol
					most_pol_loc = chunk_up_mid
					end
				end
			if surface.is_chunk_generated{math.floor(chunk_rt_mid.x/32),math.floor(chunk_rt_mid.y/32)} then
				chunk_rt_pol = surface.get_pollution(chunk_rt_mid)
				if chunk_rt_pol > most_pol then
					most_pol = chunk_rt_pol
					most_pol_loc = chunk_rt_mid
					end
				end
			if surface.is_chunk_generated{math.floor(chunk_dn_mid.x/32),math.floor(chunk_dn_mid.y/32)} then
				chunk_dn_pol = surface.get_pollution(chunk_dn_mid)
				if chunk_dn_pol > most_pol then
					most_pol = chunk_dn_pol
					most_pol_loc = chunk_dn_mid
					end
				end
			if surface.is_chunk_generated{math.floor(chunk_lt_mid.x/32),math.floor(chunk_lt_mid.y/32)} then
				chunk_lt_pol = surface.get_pollution(chunk_lt_mid)
				if chunk_lt_pol > most_pol then
					most_pol = chunk_lt_pol
					most_pol_loc = chunk_lt_mid
					end
				end

			-- so: check all orthogonal cells pollution, pick the highest of them and current cell, this is the possible "move" (or attack, if current)
			-- IF it is the current cell then attack, no more checks to do
			-- IF not then check if there are any player units in the chunk
			-- IF no player units in chunk, then move
			-- IF player units in chunk, compare weight of pollution difference vs weight of number of units and decide move vs attack based on this weighting
			local enemies = 0
			local attack = false
			local attack_loc = position
			local toughest_target = 0
			local move = false
			for key, targets in pairs(surface.find_entities_filtered({area={{position.x - 16, position.y - 16}, {position.x + 16, position.y + 16}}, force= game.forces.player})) do
				if targets.health ~= nil then
					if targets.health > toughest_target then
						toughest_target = targets.health
						attack_loc = targets.position
						end
					enemies = enemies + targets.health
					end
				end

			if enemies > 0 then
				if most_pol <= chunk_pol then
					attack = true
				elseif (most_pol - chunk_pol) * worm_pollution_weight < enemies then
					attack = true
				else
					move = true
					end
			elseif most_pol <= chunk_pol then
				attack = false
				move = false
			else
				move = true
				end
			if attack == true then
				local attacking_worm = surface.create_entity{name = attack_type, position = {x = attack_loc.x + worm_attacking_xoffset, y = attack_loc.y + worm_attacking_yoffset}}
				table.insert(global.attacking_worms, {worm=attacking_worm,attacks=attack_endurance}) 
				Quake_Area(attack_loc,2)
				table.remove(global.burrowing_worms, i)
				burrowing_worms.destroy()
--				game.player.print("Attack!")
			elseif move == true then
				Quake_Area(position,2)
				Quake_Area(most_pol_loc,2)
				burrowing_worms.teleport(most_pol_loc)
--				game.player.print("Move!")
				end
			end
		end
	end

-- run attack logic, pick an attack point, cause quakes and damaging explosions on that point, when the point is clear surface the worm! (at which point this script is done and standard AI takes over)
function check_attacking_worms ()
	local surface = game.surfaces["nauvis"]
	if global.attacking_worms == nil then global.attacking_worms = {} end
	for i, wormattacks in pairs(global.attacking_worms) do
		if wormattacks.worm then
			local worm = wormattacks.worm
			if worm.valid then
				local XY = {x=0,y=0}
				local worm_turret = ""
				local worm_attack = ""
				local worm_surface = ""
				if worm.name == "behemoth-worm-attacking" then
					worm_turret = "behemoth-worm-turret"
					worm_attack = "behemoth-worm-breakthrough"
					worm_surface = "behemoth-worm-surface"
				elseif worm.name == "colossal-worm-attacking" then
					worm_turret = "colossal-worm-turret"
					worm_attack = "colossal-worm-breakthrough"
					worm_surface = "colossal-worm-surface"
					end
				XY = {x = worm.position.x - worm_attacking_xoffset, y = worm.position.y - worm_attacking_yoffset}
				Quake_Area(XY,2)
				if surface.can_place_entity{name = worm_turret, position = XY} then
					table.remove(global.attacking_worms,i)
					worm.destroy()
					surface.create_entity{name = worm_surface, position = XY, force= game.forces.enemy}
					surface.create_entity{name = worm_turret, position = XY, force= game.forces.enemy}
--					game.player.print("Breach!")
				elseif wormattacks.attacks > 0 then
					surface.create_entity{name = worm_attack, position = XY, force= game.forces.enemy}
					wormattacks.attacks = wormattacks.attacks - 1
--					game.player.print("Thud!")
				else
					table.remove(global.attacking_worms,i)
					worm.destroy()
--					game.player.print("Die!")
					surface.create_entity{name = "worm-dying-underground", position = XY}
					--die!
					end
				end
			end
		end
	end

function check_worms ()
	if game.tick % worm_attack_update_ticks == 0 then
		check_attacking_worms ()
		end
	if game.tick % worm_burrow_update_ticks == 0 then
		check_burrowing_worms ()
		end
	if game.tick % worm_evolve_update_ticks == 0 then
		find_worms ()
		end
	if game.tick % worm_evolve_update_ticks == math.floor(worm_evolve_update_ticks / 10) then
		check_worm_evolution ()
		end
	if game.tick % worm_evolve_update_ticks == math.floor(worm_evolve_update_ticks / 5) then
		check_worm_burrow ()
		end
	end
	
script.on_event(defines.events.on_tick, function(event)
	check_worms()
end)