--require "prototypes.entity.demo-worm-animations"
--require "prototypes.entity.demo-enemy-sounds"
--require "util"

behemoth_worm_scale = 1.5
behemoth_worm_tint = {r=0.64, g=0.88, b=0.90, a=1.0}

colossal_worm_scale = 2.0
colossal_worm_tint = {r=0.94, g=0.98, b=0.90, a=1.0}

function shift_behemoth_worm(shiftx, shifty)
  return {shiftx - 0.3, shifty + 0.3}
end

function shift_colossal_worm(shiftx, shifty)
  return {shiftx - 0.4, shifty + 0.4}
end

data:extend(
{
  {
    type = "turret",
    name = "behemoth-worm-turret",
    icon = "__base__/graphics/icons/big-worm.png",
    flags = {"placeable-player", "placeable-enemy", "not-repairable", "breaths-air"},
    max_health = 750,
    order="b-b-f",
    subgroup="enemies",
    resistances =
    {
      {
        type = "physical",
        decrease = 10,
      },
      {
        type = "explosion",
        decrease = 15,
        percent = 30,
      },
      {
        type = "acid",
        decrease = 20,
        percent = 50,
      }
    },
    healing_per_tick = 0.02,
    collision_box = {{-1.9, -1.4}, {1.9, 1.4}},
    selection_box = {{-1.9, -1.4}, {1.9, 1.4}},
    shooting_cursor_size = 4,
    rotation_speed = 1,
    corpse = "behemoth-worm-corpse",
    dying_explosion = "blood-explosion-big",
    dying_sound = make_worm_dying_sounds(1.0),
    inventory_size = 2,
    folded_speed = 0.01,
    folded_animation = worm_folded_animation(behemoth_worm_scale, behemoth_worm_tint),
--    prepare_range = 25,
    preparing_speed = 0.025,
    preparing_animation = worm_preparing_animation(behemoth_worm_scale, behemoth_worm_tint, "forward"),
    prepared_speed = 0.015,
    prepared_animation = worm_prepared_animation(behemoth_worm_scale, behemoth_worm_tint),
    starting_attack_speed = 0.03,
    starting_attack_animation = worm_attack_animation(behemoth_worm_scale, behemoth_worm_tint, "forward"),
    starting_attack_sound = make_worm_roars(0.95),
    ending_attack_speed = 0.03,
    ending_attack_animation = worm_attack_animation(behemoth_worm_scale, behemoth_worm_tint, "backward"),
    folding_speed = 0.015,
    folding_animation =  worm_preparing_animation(behemoth_worm_scale, behemoth_worm_tint, "backward"),
    prepare_range = 35,
    attack_parameters =
    {
      type = "projectile",
      ammo_category = "rocket",
      cooldown = 120,
      range = 30,--25,
      projectile_creation_distance = 2.1,
      damage_modifier = 8,
      ammo_type =
      {
		category = "biological",
		target_type = "direction",
		source_effects =
		{
			type = "create-explosion",
			entity_name = "acid-splash-purple"
		},
		action =
		{
          type = "direct",
          action_delivery =
          {
			type = "projectile",
			projectile = "medium-acid-splatball-purple",
			direction_deviation = 0.1,
			range_deviation = 0.2,
			max_range = 30,
            starting_speed = 1.1
          }
        }
      },
    }
  },
  {
    type = "turret",
    name = "colossal-worm-turret",
    icon = "__base__/graphics/icons/big-worm.png",
    flags = {"placeable-player", "placeable-enemy", "not-repairable", "breaths-air"},
    max_health = 1000,
    order="b-b-f",
    subgroup="enemies",
    resistances =
    {
      {
        type = "physical",
        decrease = 20,
      },
      {
        type = "explosion",
        decrease = 50,
        percent = 30,
      },
      {
        type = "acid",
        decrease = 20,
        percent = 50,
      }
    },
    healing_per_tick = 0.02,
    collision_box = {{-2.9, -2.5}, {2.9, 2.5}},
    selection_box = {{-2.9, -2.5}, {2.9, 2.5}},
    shooting_cursor_size = 4,
    rotation_speed = 1,
    corpse = "colossal-worm-corpse",
    dying_explosion = "blood-explosion-big",
    dying_sound = make_worm_dying_sounds(1.0),
    inventory_size = 2,
    folded_speed = 0.01,
    folded_animation = worm_folded_animation(colossal_worm_scale, colossal_worm_tint),
--    prepare_range = 25,
    preparing_speed = 0.025,
    preparing_animation = worm_preparing_animation(colossal_worm_scale, colossal_worm_tint, "forward"),
    prepared_speed = 0.015,
    prepared_animation = worm_prepared_animation(colossal_worm_scale, colossal_worm_tint),
    starting_attack_speed = 0.03,
    starting_attack_animation = worm_attack_animation(colossal_worm_scale, colossal_worm_tint, "forward"),
    starting_attack_sound = make_worm_roars(0.95),
    ending_attack_speed = 0.03,
    ending_attack_animation = worm_attack_animation(colossal_worm_scale, colossal_worm_tint, "backward"),
    folding_speed = 0.015,
    folding_animation =  worm_preparing_animation(colossal_worm_scale, colossal_worm_tint, "backward"),
    prepare_range = 40,
    attack_parameters =
    {
      type = "projectile",
      ammo_category = "rocket",
      cooldown = 120,
      range = 35,
      projectile_creation_distance = 2.1,
      damage_modifier = 12,
      ammo_type =
      {
		category = "biological",
		target_type = "direction",
		source_effects =
		{
			type = "create-explosion",
			entity_name = "acid-splash-purple"
		},
		action =
		{
			type = "direct",
			action_delivery =
			{
				type = "projectile",
				projectile = "huge-acid-splatball-purple",
				direction_deviation = 0.1,
				range_deviation = 0.2,
				max_range = 30,
				starting_speed = 1.1
			}
		}
      }
    }
  },
  {
    type = "corpse",
    name = "behemoth-worm-corpse",
    icon = "__base__/graphics/icons/big-worm-corpse.png",
    selection_box = {{-0.8, -0.8}, {0.8, 0.8}},
    selectable_in_game = false,
    subgroup="corpses",
    order = "c[corpse]-c[worm]-c[behemoth]",
    flags = {"placeable-neutral", "placeable-off-grid", "building-direction-8-way", "not-repairable", "not-on-map"},
    dying_speed = 0.01,
    time_before_removed = 15 * 60 * 60,
    final_render_layer = "corpse",
    animation = worm_die_animation(behemoth_worm_scale, behemoth_worm_tint)
  },
  {
    type = "corpse",
    name = "colossal-worm-corpse",
    icon = "__base__/graphics/icons/big-worm-corpse.png",
    selection_box = {{-0.8, -0.8}, {0.8, 0.8}},
    selectable_in_game = false,
    subgroup="corpses",
    order = "c[corpse]-c[worm]-c[colossal]",
    flags = {"placeable-neutral", "placeable-off-grid", "building-direction-8-way", "not-repairable", "not-on-map"},
    dying_speed = 0.01,
    time_before_removed = 15 * 60 * 60,
    final_render_layer = "corpse",
    animation = worm_die_animation(colossal_worm_scale, colossal_worm_tint)
  }
}
)
